<?php
require_once __DIR__ .'/log4php/Logger.php';
require_once __DIR__ .'/omsservice/SubscriptionBillingClientService.php';
require_once __DIR__ .'/billingAlgorithmProcessors/BillingAlgorithmProcessor.php';
require_once __DIR__ .'/billingAlgorithmProcessors/QuotedFeeProcessor.php';
require_once __DIR__ .'/billingAlgorithmProcessors/FixedFeeProcessor.php';

class SubscriptionBilling {
	private $logger;
	private $omsWebService;
	
	function __construct($options) {
		if(!is_object($options)) {
			throw new InvalidArgumentException("options not specified");
		}
		
		if(!isset($options->username) || !is_string($options->username) || empty($options->username)) {
			throw new InvalidArgumentException("username not specified");
		}
		
		if(!isset($options->password) || !is_string($options->password) || empty($options->password)) {
			throw new InvalidArgumentException("password not specified");
		}
		
		if(!isset($options->url) || !is_string($options->url) || empty($options->url)) {
			throw new InvalidArgumentException("url not specified");
		}
		
		
		// Initialize the logger
		$this->logger = Logger::getLogger(__CLASS__);
		
		if($this->logger->isDebugEnabled()) {
			$this->logger->debug('[serviceUrl:'.$options->url.'][username:'.$options->username.'][password:'.str_repeat('*', strlen($options->password)).']');
		}
		
		// Connect to the oms soap api		
		ini_set("soap.wsdl_cache_enabled", "0"); // disabling WSDL cache
		
		$this->omsWebService = new \omsservice\SubscriptionBillingClientService(
			array(
				'login' => $options->username,
				'password' => $options->password,
				'location' => $options->url,
				'trace' => 1,
				'exceptions' => 1,
				'features' => SOAP_SINGLE_ELEMENT_ARRAYS // + SOAP_USE_XSI_ARRAY_TYPE
			),
			__DIR__ .'/omsservice/SubscriptionBillingClientService.wsdl');
		
		$this->checkService();
	}
	
	/**
	 * 
	 * Bulk generate new subscription period(s)
	 * @param string $administrationUnitId
	 * @param DateTime $billingDate
	 * @param string $billingAlgorithmId
	 * @throws Exception
	 */
	function createSubscriptionPeriods(/*string*/ $administrationUnitId, DateTime $billingDate, /*string*/ $billingAlgorithmId) {
		
		if(!isset($billingDate)) {
			$billingDate = new DateTime(); // current time
		}
		
		if($this->logger->isInfoEnabled()) {
			$this->logger->info(
			'[administrationUnitId:'.(isset($administrationUnitId) ? $administrationUnitId : null).']'.
			'[billingDate:'.$billingDate->format(DateTime::W3C).']'.
			'[billingAlgorithmId:'.(isset($billingAlgorithmId) ? $billingAlgorithmId : null).']');
		}

		// Initialize the processor that can calculate the potential subscription periods		
		$billingAlgorithmProcessor = $this->initializeBillingAlgorithmProcessor($billingAlgorithmId);
		
		// Array of subscription renewal ids that haven't been billed for the effective date
		$subscriptionRenewalIds = $this->getUnBilledSubscriptionRenewalIds($administrationUnitId, $billingAlgorithmId, $billingDate);
	
		if(!empty($subscriptionRenewalIds)) {	
			foreach ($subscriptionRenewalIds as &$subscriptionRenewalId) {
				try {
					// Get the billing context for the subscription renewal and billing date
					$subscriptionRenewalBillingContext = $this->getSubscriptionRenewalBillingContext($subscriptionRenewalId, $billingDate);
					// Calculate the potential subscription period using the billing algorithm processor					
					$subscriptionPeriodCreateRequestTO = $billingAlgorithmProcessor->calculatePotentialSubscriptionPeriod($subscriptionRenewalBillingContext);
						
					// Submit the subscription period
					$subscriptionPeriodId = $this->submitSubscriptionPeriod($subscriptionPeriodCreateRequestTO);
				} catch(Exception $e) {
					$this->logger->error('Cannot create subscription period '.
					'[subscriptionRenewalId:'.(isset($subscriptionRenewalId) ? $subscriptionRenewalId : null).']'.
					'[effectiveDate:'.(isset($billingDate) ? $billingDate->format(DateTime::W3C) : null).']', $e);
				}
			}
		} else {
			if($this->logger->isWarnEnabled()) {
				$this->logger->warn("No unbilled subscription renewals found");
			}
		}
	}	
	
	/**
	*
	* Generate new subscription period
	* @param string $administrationUnitId
	* @param string $billingAlgorithm
	* @param DateTime $effectiveDate
	* @return string
	* @throws Exception
	*/
	function createSubscriptionPeriod(/*string*/ $associationId, DateTime $billingDate) {
	
		if(!isset($billingDate)) {
			$billingDate = new DateTime(); // current time
		}
	
		if($this->logger->isInfoEnabled()) {
			$this->logger->info(
				'[associationId:'.(isset($associationId) ? $associationId : null).']'.
				'[billingDate:'.(isset($billingDate) ? $billingDate->format(DateTime::W3C) : null).']');
		}
		
		// Get the subscription renewal for the association and billing date
		$subscriptionRenewal = $this->getAssociationSubscriptionRenewal($associationId, $billingDate);
		
		// Get the billing context for the subscription renewal and billing date
		$subscriptionRenewalBillingContext = $this->getSubscriptionRenewalBillingContext($subscriptionRenewal->id, $billingDate);
				
		// Initialize the billing algorithm processor that can calculate the potential subscription period
		$billingAlgorithmProcessor = $this->initializeBillingAlgorithmProcessor($subscriptionRenewalBillingContext->subscriptionMethod->billingAlgorithmId);

		// Calculate the potential subscription period using the billing algorithm processor
		$subscriptionPeriodCreateRequestTO = $billingAlgorithmProcessor->calculatePotentialSubscriptionPeriod($subscriptionRenewalBillingContext);
		
		// Submit the subscription period
		$subscriptionPeriodId = $this->submitSubscriptionPeriod($subscriptionPeriodCreateRequestTO);
		return $subscriptionPeriodId;
	}
	
	/**
	* Check whether the oms soap service is available or not
	* @throws Exception
	*/
	private function checkService() {
		try {
			$response = $this->omsWebService->echoTest(new \omsservice\echoTest("SubscriptionBilling"));
			
			if($this->logger->isTraceEnabled()) {
				$this->logger->trace("[soap request:\n".$this->omsWebService->__getLastRequest()."\n]\n[soap response:\n".$this->omsWebService->__getLastResponse()."]");
			}
							
			if(!isset($response) || !isset($response->return)) {
				throw new Exception("Cannot check oms service: invalid response");
			}
			
			//if($this->logger->isDebugEnabled()) {
			//	$this->logger->debug("[interfaceVersion:".$response->return."][version:".$response->return->version."][authenticatedUser:".$response->return->authenticatedUser->userName."]");
			//}
		} catch(SoapFault $e) {
			throw $this->omsSoapFault($e);
		}
	}
	
	/**
	* Initialize a billing algorithm processor for the specified billing algorithm id
	* @param string $billingAlgorithmId
	* @return BillingAlgorithmProcessor
	* @throws Exception
	*/
	private function initializeBillingAlgorithmProcessor(/*string*/ $billingAlgorithmId) {
		if($this->logger->isDebugEnabled()) {
			$this->logger->debug('[billingAlgorithmId:'.$billingAlgorithmId.']');
		}
	
		$billingAlgorithmProcessor = null;
		// Initialize the billing algorithm processor
		if(!isset($billingAlgorithmId) ) {
			throw new Exception("Cannot initialize billing algorithm processor: billing algorithm id not specified");
		} else if($billingAlgorithmId == 'FIXED_FEE') {
			$billingAlgorithmProcessor = new FixedFeeProcessor($this->omsWebService);
		} else if($billingAlgorithmId == 'QUOTED_FEE') {
			$billingAlgorithmProcessor = new QuotedFeeProcessor($this->omsWebService);
		} else {
			throw new Exception("Cannot initialize billing algorithm processor: unknown billing algorithm id '".$billingAlgorithmId."'");
		}
		
		return $billingAlgorithmProcessor;
	}
	
	/**
	 * Get the subscription renewal for the association and effective date
	 * @param string $associationId
	 * @param DateTime $effectiveDate
	 * @return \omsservice\SubscriptionRenewalTO
	 * @throws Exception
	 */
	private function getAssociationSubscriptionRenewal(/*string*/ $associationId, DateTime $effectiveDate) {
		if($this->logger->isDebugEnabled()) {
			$this->logger->debug(
					'[associationId:'.(isset($associationId) ? $associationId : null).']'.
					'[effectiveDate:'.(isset($effectiveDate) ? $effectiveDate->format(DateTime::W3C) : null).']');
		}
		
		if(!isset($associationId) || !isset($effectiveDate)) {
			throw new Exception ("Cannot get association subscription renewal: search parameters not specified");
		}
		
		$subscriptionRenewals = null;
		
		try {
			$searchParameters = new \omsservice\subscriptionRenewalSearchRequestTO();
			$searchParameters->associationIds = array($associationId);			
			$searchParameters->from = $effectiveDate->format(DateTime::W3C);
			$searchParameters->to = $effectiveDate->format(DateTime::W3C);
			$searchParameters->rowLimit = 2;
			$searchParameters->administrationUnitIds = array();
			$searchParameters->groupIds = array();
			$searchParameters->ids = array();
			$searchParameters->subscriptionMethodIds = array();
			
			$searchSubscriptionRenewalsResponse = $this->omsWebService->searchSubscriptionRenewals(new \omsservice\searchSubscriptionRenewals($searchParameters));

			if($this->logger->isTraceEnabled()) {
				$this->logger->trace("[soap request:\n".$this->omsWebService->__getLastRequest()."\n]\n[soap response:\n".$this->omsWebService->__getLastResponse()."]");
			}
			
			if(!isset($searchSubscriptionRenewalsResponse) || !isset($searchSubscriptionRenewalsResponse->return)) {
				throw new Exception ("Cannot get association subscription renewal: invalid response");
			}
				
			$subscriptionRenewals = $searchSubscriptionRenewalsResponse->return;
		}catch (SoapFault $soapFault) {
			throw $this->omsSoapFault($soapFault);
		}
		
		if(!is_array($subscriptionRenewals) || count($subscriptionRenewals) !== 1) {
			throw new Exception ("Cannot get association subscription renewal: single subscription renewal not found");
		}
		
		return $subscriptionRenewals[0];
	}
	
	/**
	 * 
	 * Get a list of unbilled subscription renewal ids
	 * @param string $administrationUnitId
	 * @param string $billingAlgorithmId
	 * @param DateTime $effectiveDate
	 * @return string[]
	 * @throws Exception
	 */
	private function getUnbilledSubscriptionRenewalIds(/*string*/ $administrationUnitId, /*string*/ $billingAlgorithmId, DateTime $effectiveDate) {
		if($this->logger->isDebugEnabled()) {
			$this->logger->debug(
					'[administrationUnitId:'.(isset($administrationUnitId) ? $administrationUnitId : null).']'.
					'[billingAlgorithmId:'.(isset($billingAlgorithmId) ? $billingAlgorithmId : null).']'.
					'[effectiveDate:'.(isset($effectiveDate) ? $effectiveDate->format(DateTime::W3C) : null).']');
		}
		
		try {
			$response = $this->omsWebService->getUnBilledSubscriptionRenewalIds(new \omsservice\getUnBilledSubscriptionRenewalIds($administrationUnitId, $billingAlgorithmId, $effectiveDate->format(DateTime::W3C)));
				
			if($this->logger->isTraceEnabled()) {
				$this->logger->trace("[soap request:\n".$this->omsWebService->__getLastRequest()."\n]\n[soap response:\n".$this->omsWebService->__getLastResponse()."]");
			}
							
			if(!isset($response)) {
				throw new Exception("Cannot get unbilled subscription renewal ids: invalid response");
			}
			
			return isset($response->return) ? $response->return : array();
		} catch(SoapFault $soapFault) {
			throw $this->omsSoapFault($soapFault);
		}
	}
		
	/**
	 * 
	 * Get the billing context for the subscription renewal and billing date
	 * @param string $subscriptionRenewalId
	 * @param DateTime $billingDate
	 * @throws Exception
	 * @return \omsservice\subscriptionRenewalPlusBillingContextTO
	 */
	private function getSubscriptionRenewalBillingContext(/*string*/ $subscriptionRenewalId, DateTime $billingDate) {
		if($this->logger->isDebugEnabled()) {
			$this->logger->debug(
						'[subscriptionRenewalId:'.(isset($subscriptionRenewalId) ? $subscriptionRenewalId : null).']'.
						'[billingDate:'.(isset($billingDate) ? $billingDate->format(DateTime::W3C) : null).']');
		}
		
		if(!isset($subscriptionRenewalId)) {
			throw new Exception("Cannot get subscription renewal billing context: subscription renewal id not specified");
		}
		
		if(!isset($billingDate)) {
			throw new Exception("Cannot get subscription renewal billing context: billing effective date not specified");
		}
	
		$subscriptionRenewalBillingContext = null;
			
		//Get the subscription renewal billing context details
		try {
			$getSubscriptionRenewalBillingContextResponse = $this->omsWebService->getSubscriptionRenewalBillingContext(new \omsservice\getSubscriptionRenewalBillingContext($subscriptionRenewalId, $billingDate->format(DateTime::W3C)));

			if($this->logger->isTraceEnabled()) {
				$this->logger->trace("[soap request:\n".$this->omsWebService->__getLastRequest()."\n]\n[soap response:\n".$this->omsWebService->__getLastResponse()."]");
			}

			if(!isset($getSubscriptionRenewalBillingContextResponse) || !isset($getSubscriptionRenewalBillingContextResponse->return)) {
				throw new Exception("Cannot get subscription renewal billing context: invalid response");
			}
				
			$subscriptionRenewalBillingContext = $getSubscriptionRenewalBillingContextResponse->return;
		} catch(SoapFault $soapFault) {
			throw $this->omsSoapFault($soapFault);
		}
	
		return $subscriptionRenewalBillingContext;
	}
	
	/**
	 * 
	 * Submit the subscription period to the oms server
	 * @param \omsservice\subscriptionPeriodCreateRequestTO $subscriptionPeriodCreateRequestTO
	 * @return string
	 * @throws Exception
	 */
	private function submitSubscriptionPeriod(\omsservice\subscriptionPeriodCreateRequestTO $subscriptionPeriodCreateRequestTO) {
		if(!isset($subscriptionPeriodCreateRequestTO)) {
			throw new Exception("Cannot submit subscription period: subscription period not specified");
		}
		
		//var_dump($subscriptionPeriodCreateRequestTO);
		//throw new Exception("defanged for testing");
		
		if($this->logger->isInfoEnabled()) {
			$this->logger->info('submitting subscription period '.
			'[associationId:'.(isset($subscriptionPeriodCreateRequestTO->associationId) ? $subscriptionPeriodCreateRequestTO->associationId : null).']'.
			'[subscriptionMethodId:'.(isset($subscriptionPeriodCreateRequestTO->subscriptionMethodId) ? $subscriptionPeriodCreateRequestTO->subscriptionMethodId : null).']'.
			'[from:'.(isset($subscriptionPeriodCreateRequestTO->from) ? $subscriptionPeriodCreateRequestTO->from : null).']'.
			'[to:'.(isset($subscriptionPeriodCreateRequestTO->to) ? $subscriptionPeriodCreateRequestTO->to : null).']'.
			'[fee:'.(isset($subscriptionPeriodCreateRequestTO->fee) ? $subscriptionPeriodCreateRequestTO->fee : null).']'.
			'[gst:'.(isset($subscriptionPeriodCreateRequestTO->gst) ? $subscriptionPeriodCreateRequestTO->gst : null).']'.
			'[description:'.(isset($subscriptionPeriodCreateRequestTO->description) ? $subscriptionPeriodCreateRequestTO->description : null).']');
		}
				
		$subscriptionPeriodId = null;
			
		try {
			$createSubscriptionPeriodResponse = $this->omsWebService->createSubscriptionPeriod(new \omsservice\createSubscriptionPeriod($subscriptionPeriodCreateRequestTO));
		
			if($this->logger->isTraceEnabled()) {
				$this->logger->trace("[soap request:\n".$this->omsWebService->__getLastRequest()."\n]\n[soap response:\n".$this->omsWebService->__getLastResponse()."]");
			}
		
			if(!isset($createSubscriptionPeriodResponse) || !isset($createSubscriptionPeriodResponse->return)) {
				throw new Exception("Cannot submit subscription period: invalid response");
			}
		
			$subscriptionPeriodId = $createSubscriptionPeriodResponse->return;
		} catch(SoapFault $soapFault) {
			throw $this->omsSoapFault($soapFault);
		}
		
		if($this->logger->isDebugEnabled()) {
			$this->logger->debug('submitted subscription period '.
			'[subscriptionPeriodId:'.(isset($subscriptionPeriodId) ? $subscriptionPeriodId : null).']'.
			'[associationId:'.(isset($subscriptionPeriodCreateRequestTO->associationId) ? $subscriptionPeriodCreateRequestTO->associationId : null).']'.
			'[from:'.(isset($subscriptionPeriodCreateRequestTO->from) ? $subscriptionPeriodCreateRequestTO->from : null).']'.
			'[to:'.(isset($subscriptionPeriodCreateRequestTO->to) ? $subscriptionPeriodCreateRequestTO->to : null).']'.
			'[fee:'.(isset($subscriptionPeriodCreateRequestTO->fee) ? $subscriptionPeriodCreateRequestTO->fee : null).']'.
			'[gst:'.(isset($subscriptionPeriodCreateRequestTO->gst) ? $subscriptionPeriodCreateRequestTO->gst : null).']'.
			'[description:'.(isset($subscriptionPeriodCreateRequestTO->description) ? $subscriptionPeriodCreateRequestTO->description : null).']');
		}
		
		return $subscriptionPeriodId;
	}
	
	/**
	* Wrap the soap fault in a more descriptive exception
	* @param SoapFault $soapFault
	* @return Exception
	*/
	private function omsSoapFault(SoapFault $soapFault) {
		if($this->logger->isDebugEnabled()) {
			$this->logger->debug("[soap request:\n".$this->omsWebService->__getLastRequest()."\n]\n[soap response:\n".$this->omsWebService->__getLastResponse()."]", $soapFault);
		}
	
		$returnValue = null;
	
		if(isset($soapFault->detail) && isset($soapFault->detail->ApplicationException)) {
			$errorMessage = $soapFault->getMessage();
	
			if(isset($soapFault->detail->ApplicationException->errorCode)) {
				$errorMessage .= ", " . $soapFault->detail->ApplicationException->errorCode;
	
				if(isset($soapFault->detail->ApplicationException->referenceId)) {
					$errorMessage .= "(" . $soapFault->detail->ApplicationException->referenceId . ")";
				}
			}
	
			$returnValue = new Exception($errorMessage, null, $soapFault);
		} else {
			$returnValue = $soapFault;
		}
	
		return $returnValue;
	}
	
	/**
	 * 
	 * Convert and throw a php error as an ErrorException
	 * @param unknown_type $errno
	 * @param unknown_type $errstr
	 * @param unknown_type $errfile
	 * @param unknown_type $errline
	 * @throws ErrorException
	 */
	static function exceptionErrorHandler($errno, $errstr, $errfile, $errline ) {
		throw new ErrorException($errstr, 0, $errno, $errfile, $errline);
	}
}
?>