#!/bin/sh -u
# Initiate the EFT sync for Oms Services using Paystream
#
# Copyright (c) 2004-2015 Adrian Butterworth
# All rights reserved.
#
# Author: Adrian Butterworth
#
# last modified: 2016-02-03
#
#----------------------------------------------------
#
usage() {
	cat >&2 <<-EOF

	Initiate bulk billing of subscription renewals
	by calling bill.php for each configuration file
	in the config directory

	usage: $SCRIPTNAME [switches] [confdir [YYYY-MM-DD]]
	switches:
	    -j NUMBER  [j]ob concurrency (default 1)
	    -l NUMBER  [l]og level 0 none, 1 err, 2 info, 3 debug
	    -q         [q]uiet no console output
	    -v         [v]erbose console output - multiple increases level
	    -?|h       this help
	parameters:
	   confdir     contains each organisations config file in json format
	               default: ~/conf/subscriptionbilling
	   YYYY-MM-DD  renewal date, default tomorrow

	example:
	$SCRIPTNAME -v -l2 -j3 ~/conf/subscriptionbilling  2015-03-27
	$SCRIPTNAME ~/conf/subscriptionbilling
	$SCRIPTNAME 

	EOF
}
msg () {
  case "$1" in err) L=1;; info) L=2;; debug) L=3;; *) L=99;; esac
  case "$L" in 1|2|3) LEVEL=$1; shift ;; *) LEVEL="info"; L=2 ;; esac
  test -z "$*" && return 0
  test ${VERBOSITY:=1} -ge $L && echo "${*}" >&2
  test ${LOGLEVEL:=2} -ge $L && logger -p user.$LEVEL -t $SCRIPTNAME -- "${*}"
  return 0
  # LOG_EMERG      system is unusable
  # LOG_ALERT      action must be taken immediately
  # LOG_CRIT       critical conditions
  # LOG_ERR        error conditions
  # LOG_WARNING    warning conditions
  # LOG_NOTICE     normal, but significant, condition
  # LOG_INFO       informational message
  # LOG_DEBUG      debug-level message
}
my_TMPDIR () { 
  test $# -eq 2 && {
    test "$2" == "remove" || return 1
    test -d "$1" || return
    rm -f "$1/PID" 2>/dev/null
    rmdir "$1" 2>/dev/null || msg err "remove failed - '$1' not empty" >&2
    return
  }
  export TMPDIR=${1:-${TMPDIR-"/tmp"}/$(basename $0)}
  test -f "$TMPDIR/PID" \
  && test -d /proc/ -a ! -d "/proc/$(<$TMPDIR/PID)" \
  && my_TMPDIR $TMPDIR remove
  (umask 077 && mkdir "$TMPDIR") &>/dev/null || { msg err "Abort - locked by existing '$TMPDIR'" >&2 ; exit 1; }
  echo $$ > $TMPDIR/PID
  trap "EXITSTATUS=$?; my_TMPDIR $TMPDIR remove && exit ${EXITSTATUS-0}" INT TERM EXIT;
}
#
#---------------------------------------------------
# hard env pre sets & unsets
SCRIPTNAME="${0##*/}"
SCRIPTDIR="$(cd "${0%/*}" 2>/dev/null; pwd )" # absolute path
CONFDIR="${SCRIPTDIR%/*}/etc"
SCRIPTCONF="$CONFDIR/${SCRIPTNAME%.*}.conf"
# config file settings
test -f "$SCRIPTCONF" && . $SCRIPTCONF # || { msg err "missing config file"; exit 1; }
# switch settings
while getopts l:j:qrvh? opt ; do
    case $opt in
    l)  LOGLEVEL="$OPTARG" ;;
    q)  VERBOSITY=0 ;;
    v)  VERBOSITY=$(( ${VERBOSITY:-1} +1 )) ;;
    j)  JOBSMAX="$OPTARG" ;;
    *)  usage; exit 1;;
    esac
done
shift $(( $OPTIND -1 ))
: ${JOBSMAX:=1}
test $# -le 2  || { usage;  exit 1; }
appconfdir="${1:-~/conf/subscriptionbilling}"
billdate="${2:-$(date -d tomorrow +%Y-%m-%d)}"
# create private TMPDIR that doubles as a lock
my_TMPDIR  /var/tmp/$SCRIPTNAME  # (use $SCRIPTNAME-$$ if you don't want locks)
#----------------------------------------------------

eval cd "$appconfdir" || { msg err "config directory '$appconfdir' not available";  exit 1; }

count(){ echo $# ; }
childcount(){ count $( pgrep -P$$ ) ; }
waitforjobslot(){
  while [ $(childcount) -gt $JOBSMAX ] ; do sleep 5 ; done
}

billsubscriptionrenewals() {
  pgrep -f $1 >/dev/null && { msg err "skipping sync for $1 - already running "; return; }
  msg initiating bill.php for $1 $2
  $SCRIPTDIR/bill.php $1 $2
  msg finished billing for $1 $2
}

msg start
for CF in $(ls *.json 2>/dev/null) ; do
  waitforjobslot
  { billsubscriptionrenewals "$CF" $billdate & }
done

wait
msg finish
#----------------------------------------------------
