<?php
require_once __DIR__ .'/BillingAlgorithmProcessor.php';

class QuotedFeeProcessor extends BillingAlgorithmProcessor {
	
	function __construct(\omsservice\SubscriptionBillingClientService $omsWebService) {
		parent::__construct($omsWebService);
	}
	
	public function calculatePotentialSubscriptionPeriod(\omsservice\subscriptionRenewalPlusBillingContextTO $subscriptionRenewalBillingContext) {
		if(!isset($subscriptionRenewalBillingContext)) {
			throw new Exception("Cannot calculate potential subscription period: billing context not specified");
		}
		
		if($this->logger->isDebugEnabled()) {
			$this->logger->debug('subscription renewal billing context'.
			'[personId:'.(isset($subscriptionRenewalBillingContext->person) ? $subscriptionRenewalBillingContext->person->id : null) .']'.
			'[personName:'.(isset($subscriptionRenewalBillingContext->person) ? $subscriptionRenewalBillingContext->person->firstNames . ' ' . $subscriptionRenewalBillingContext->person->lastName : null) .']'.
			'[billingDate:'.(isset($subscriptionRenewalBillingContext->billingDate) ? $subscriptionRenewalBillingContext->billingDate : null) .']'.
			'[overlappingSubscriptionPeriodId:'.(isset($subscriptionRenewalBillingContext->overlappingSubscriptionPeriodId) ? $subscriptionRenewalBillingContext->overlappingSubscriptionPeriodId : null) .']'.
			'[subscriptionPeriodPreferredDays:'.(isset($subscriptionRenewalBillingContext->subscriptionPeriodPreferredDays) ? $subscriptionRenewalBillingContext->subscriptionPeriodPreferredDays : null) .']'.				
			'[subscriptionPeriodPreferredFrom:'.(isset($subscriptionRenewalBillingContext->subscriptionPeriodPreferredFrom) ? $subscriptionRenewalBillingContext->subscriptionPeriodPreferredFrom : null) .']'.				
			'[subscriptionPeriodPreferredTo:'.(isset($subscriptionRenewalBillingContext->subscriptionPeriodPreferredTo) ? $subscriptionRenewalBillingContext->subscriptionPeriodPreferredTo : null) .']'.				
			'[subscriptionPeriodAllowedDays:'.(isset($subscriptionRenewalBillingContext->subscriptionPeriodAllowedDays) ? $subscriptionRenewalBillingContext->subscriptionPeriodAllowedDays : null) .']'.				
			'[subscriptionPeriodAllowedFrom:'.(isset($subscriptionRenewalBillingContext->subscriptionPeriodAllowedFrom) ? $subscriptionRenewalBillingContext->subscriptionPeriodAllowedFrom : null) .']'.				
			'[subscriptionPeriodAllowedTo:'.(isset($subscriptionRenewalBillingContext->subscriptionPeriodAllowedTo) ? $subscriptionRenewalBillingContext->subscriptionPeriodAllowedTo : null) .']'.
			'[subscriptionRenewalId:'.((isset($subscriptionRenewalBillingContext->subscriptionRenewal) && isset($subscriptionRenewalBillingContext->subscriptionRenewal->id)) ? $subscriptionRenewalBillingContext->subscriptionRenewal->id : null) .']'.
			'[subscriptionMethodId:'.((isset($subscriptionRenewalBillingContext->subscriptionMethod) && isset($subscriptionRenewalBillingContext->subscriptionMethod->id)) ? $subscriptionRenewalBillingContext->subscriptionMethod->id : null) .']'.
			'[subscriptionMethodName:'.((isset($subscriptionRenewalBillingContext->subscriptionMethod) && isset($subscriptionRenewalBillingContext->subscriptionMethod->name)) ? $subscriptionRenewalBillingContext->subscriptionMethod->name : null) .']'.
			'[subscriptionFrequencyId:'.((isset($subscriptionRenewalBillingContext->subscriptionFrequency) && isset($subscriptionRenewalBillingContext->subscriptionFrequency->id)) ? $subscriptionRenewalBillingContext->subscriptionFrequency->id : null) .']'.
			'[subscriptionFrequencyName:'.((isset($subscriptionRenewalBillingContext->subscriptionFrequency) && isset($subscriptionRenewalBillingContext->subscriptionFrequency->name)) ? $subscriptionRenewalBillingContext->subscriptionFrequency->name : null) .']');
		}

		if(isset($subscriptionRenewalBillingContext->overlappingSubscriptionPeriodId)) {
			throw new Exception("Cannot calculate potential subscription period: overlapping subscription period found (".$subscriptionRenewalBillingContext->overlappingSubscriptionPeriodId.")");
		}
		
		if(!isset($subscriptionRenewalBillingContext->subscriptionRenewal)) {
			throw new Exception("Cannot calculate potential subscription period: subscription renewal not found");
		}

		if(!isset($subscriptionRenewalBillingContext->subscriptionMethod)) {
			throw new Exception("Cannot calculate potential subscription period: subscription method not found");
		}
		
		if(!isset($subscriptionRenewalBillingContext->subscriptionPeriodPreferredFrom) || !isset($subscriptionRenewalBillingContext->subscriptionPeriodPreferredTo) || !isset($subscriptionRenewalBillingContext->subscriptionPeriodPreferredDays)) {
			throw new Exception("Cannot calculate potential subscription period: subscription preferred time period not specified");
		}
		
		if(!isset($subscriptionRenewalBillingContext->subscriptionPeriodAllowedFrom) || !isset($subscriptionRenewalBillingContext->subscriptionPeriodAllowedTo) || !isset($subscriptionRenewalBillingContext->subscriptionPeriodAllowedDays)) {
			throw new Exception("Cannot calculate potential subscription period: subscription allowable time period not specified");
		}
		
		$subscriptionQuotedFee = null;
		
		try {
			$getSubscriptionQuotedFeeResponse = $this->omsWebService->getSubscriptionQuotedFee(new \omsservice\getSubscriptionQuotedFee($subscriptionRenewalBillingContext->subscriptionRenewal->id));
		
			if($this->logger->isTraceEnabled()) {
				$this->logger->trace("[soap request:\n".$this->omsWebService->__getLastRequest()."\n]\n[soap response:\n".$this->omsWebService->__getLastResponse()."]");
			}
			
			if(!isset($getSubscriptionQuotedFeeResponse) || !isset($getSubscriptionQuotedFeeResponse->return)) {
				throw new Exception("Cannot calculate potential subscription period: invalid response");
			}
			
			$subscriptionQuotedFee = $getSubscriptionQuotedFeeResponse->return;
		} catch(SoapFault $soapFault) {
			throw new Exception("Cannot calculate potential subscription period: ".$soapFault->getMessage(), null, $soapFault);
		}
		
		if(!isset($subscriptionQuotedFee)) {
			throw new Exception("Cannot calculate potential subscription period: subscription quoted fee not found");
		}
		
		$preferredFee = $subscriptionQuotedFee->amount;
		$fee = null;
		
		if ($subscriptionRenewalBillingContext->subscriptionPeriodAllowedDays <= 0 || $subscriptionRenewalBillingContext->subscriptionPeriodPreferredDays <= 0) {
			$fee = 0;
		} else if($subscriptionRenewalBillingContext->subscriptionPeriodAllowedDays === $subscriptionRenewalBillingContext->subscriptionPeriodPreferredDays) {
			$fee = $preferredFee;
		} else {
			$allowableFeeFactor = $subscriptionRenewalBillingContext->subscriptionPeriodAllowedDays/$subscriptionRenewalBillingContext->subscriptionPeriodPreferredDays;
			$fee = (int)round($preferredFee * $allowableFeeFactor);
		}
		
		$gst = $this->calculateGst(new DateTime($subscriptionRenewalBillingContext->subscriptionPeriodAllowedFrom), $fee, true);
		$fee -= $gst;
		
		$returnValue = new \omsservice\subscriptionPeriodCreateRequestTO();
		$returnValue->associationId = $subscriptionRenewalBillingContext->subscriptionRenewal->associationId;
		$returnValue->description = "Subscription - ".$subscriptionRenewalBillingContext->person->firstNames." ".$subscriptionRenewalBillingContext->person->lastName;
		$returnValue->fee = $fee;
		$returnValue->gst = $gst;
		$returnValue->financialCostCentreId = $subscriptionRenewalBillingContext->subscriptionMethod->financialCostCentreId;
		$returnValue->from = $subscriptionRenewalBillingContext->subscriptionPeriodAllowedFrom;
		$returnValue->id = null;
		$returnValue->payableDate = $subscriptionRenewalBillingContext->subscriptionPeriodAllowedFrom;
		$returnValue->subscriberFinancialAccountId = $subscriptionRenewalBillingContext->subscriptionRenewal->financialAccountId;
		$returnValue->subscriptionFeeFinancialAccountId = $subscriptionRenewalBillingContext->subscriptionMethod->financialAccountId;
		$returnValue->subscriptionMethodId = $subscriptionRenewalBillingContext->subscriptionMethod->id;
		$returnValue->to = $subscriptionRenewalBillingContext->subscriptionPeriodAllowedTo;

		return $returnValue;
	}
	
	/**
	*
	* Get the GST factor for the specified date
	* @param DateTime $effectiveDate
	* @throws Exception
	* @return number
	*/
	private function getGstFactor(DateTime $effectiveDate) {		
		if($effectiveDate === null) {
			throw new Exception("Could not determine the gst factor: gst effective date not specified");
		}
	
		$gstEffectiveDate = DateTime::createFromFormat('!Y-m-d',$effectiveDate->format('Y-m-d'));
		$preGstDate = new DateTime('2000-06-30');
		$gstFactor;
	
		if($gstEffectiveDate > $preGstDate) {
			$gstFactor = 0.1;
		} else {
			$gstFactor = 0;
		}
	
		return $gstFactor;
	}
	
	/**
	 *
	 * Calculate the GST for an amount, the GST can be inclusive or exclusive of the amount
	 * @param DateTime $effectiveDate
	 * @param int $amount
	 * @param boolean $gstInclusive
	 */
	private function calculateGst(DateTime $effectiveDate, $amount, $gstInclusive = true) {	
		if(!isset($amount) || !is_int($amount)) {
			throw new Exception("Could not calculate gst: amount is not an integer");
		}
		
		if(!isset($gstInclusive) || !is_bool($gstInclusive)) {
			throw new Exception("Could not calculate gst: gst inclusive or exclusive not specified");
		}
		
		$gstFactor = $this->getGstFactor($effectiveDate);
		$gst;
		
		if($gstFactor === 0) {
			$gst = 0;
		} else if($gstInclusive === true) {
			$gst = (int)round($amount * $gstFactor/($gstFactor+1));
		} else if ($gstInclusive === false) {
			$gst = (int)round($amount * $gstFactor);
		}
	
		return $gst;
	}
}
?>